@@ -1,6 +1,3 @@
-from __future__ import annotations
-
-import asyncio
 import itertools
 import os
 import pathlib
@@ -73,9 +70,7 @@ from .dap_types import (
 )
 
 if get_robot_version() >= (7, 0):
-    from robot.running.invalidkeyword import (
-        InvalidKeywordRunner as UserKeywordHandler,
-    )
+    from robot.running import UserKeyword as UserKeywordHandler
 else:
     from robot.running.userkeyword import UserKeywordHandler
 
@@ -159,14 +154,14 @@ class InvalidThreadIdError(Exception):
 class StackFrameEntry:
     def __init__(
         self,
-        parent: Optional[StackFrameEntry],
+        parent: Optional["StackFrameEntry"],
         context: Any,
         name: str,
         type: str,
         source: Optional[str],
         line: Optional[int],
         column: Optional[int] = None,
-        handler: Any = None,
+        handler: Optional[UserKeywordHandler] = None,
         is_file: bool = True,
         libname: Optional[str] = None,
         kwname: Optional[str] = None,
@@ -195,7 +190,7 @@ class StackFrameEntry:
     def __repr__(self) -> str:
         return f"StackFrameEntry({self.name!r}, {self.type!r}, {self.source!r}, {self.line!r}, {self.column!r})"
 
-    def get_first_or_self(self) -> StackFrameEntry:
+    def get_first_or_self(self) -> "StackFrameEntry":
         if self.stack_frames:
             return self.stack_frames[0]
         return self
@@ -248,6 +243,12 @@ else:
         def __init__(self) -> None:
             self.steps: List[Any] = []
 
+        def start_try(self, data: "running.Try", result: "result.Try") -> None:
+            self.steps.append(data)
+
+        def end_try(self, data: "running.Try", result: "result.Try") -> None:
+            self.steps.pop()
+
         def start_keyword(self, data: running.Keyword, result: result.Keyword) -> None:
             self.steps.append(data)
 
@@ -256,14 +257,14 @@ else:
 
 
 class Debugger:
-    __instance: ClassVar[Optional[Debugger]] = None
+    __instance: ClassVar[Optional["Debugger"]] = None
     __lock: ClassVar = threading.RLock()
     __inside_instance: ClassVar = False
 
     _logger = LoggingDescriptor()
 
     @classmethod
-    def instance(cls) -> Debugger:
+    def instance(cls) -> "Debugger":
         if cls.__instance is not None:
             return cls.__instance
         with cls.__lock:
@@ -314,7 +315,6 @@ class Debugger:
         self.terminated = False
         self.attached = False
         self.path_mappings: List[PathMapping] = []
-        self.server_loop: Optional[asyncio.AbstractEventLoop] = None
 
         self._keyword_to_evaluate: Optional[Callable[..., Any]] = None
         self._evaluated_keyword_result: Any = None
@@ -500,8 +500,7 @@ class Debugger:
             self.condition.notify_all()
 
     @event
-    def send_event(sender: Any, event: Event) -> None:
-        ...
+    def send_event(sender: Any, event: Event) -> None: ...
 
     def set_breakpoints(
         self,
@@ -871,7 +870,11 @@ class Debugger:
             LOGGER.register_logger(self.debug_logger)
 
         source = attributes.get("source")
-        line_no = attributes.get("lineno", 1)
+        line_no_dummy = attributes.get("lineno", 1)
+        if isinstance(line_no_dummy, str):
+            line_no = int(line_no_dummy) if line_no_dummy else None
+        else:
+            line_no = line_no_dummy
         longname = attributes.get("longname", "")
         status = attributes.get("status", "")
         type = "SUITE"
@@ -910,14 +913,12 @@ class Debugger:
             status = attributes.get("status", "")
 
             if status == "FAIL":
-                if self.server_loop:
-                    self.process_end_state(
-                        status,
-                        {"failed_suite"},
-                        "Suite failed.",
-                        f"Suite failed{f': {v}' if (v := attributes.get('message')) else ''}",
-                    )
-
+                self.process_end_state(
+                    status,
+                    {"failed_suite"},
+                    "Suite failed.",
+                    f"Suite failed{f': {v}' if (v := attributes.get('message')) else ''}",
+                )
                 self.wait_for_running()
 
         source = attributes.get("source")
@@ -928,7 +929,11 @@ class Debugger:
 
     def start_test(self, name: str, attributes: Dict[str, Any]) -> None:
         source = attributes.get("source")
-        line_no = attributes.get("lineno", 1)
+        line_no_dummy = attributes.get("lineno", 1)
+        if isinstance(line_no_dummy, str):
+            line_no = int(line_no_dummy) if line_no_dummy else None
+        else:
+            line_no = line_no_dummy
         longname = attributes.get("longname", "")
         status = attributes.get("status", "")
 
@@ -967,10 +972,24 @@ class Debugger:
 
         self.remove_stackframe_entry(longname, type, source, line_no)
 
+    if get_robot_version() >= (7, 0):
+
+        def get_current_keyword_handler(self, name: str) -> UserKeywordHandler:
+            return EXECUTION_CONTEXTS.current.namespace.get_runner(name).keyword
+
+    else:
+
+        def get_current_keyword_handler(self, name: str) -> UserKeywordHandler:
+            return EXECUTION_CONTEXTS.current.namespace.get_runner(name)._handler
+
     def start_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
         status = attributes.get("status", "")
         source = attributes.get("source")
-        line_no = attributes.get("lineno")
+        line_no_dummy = attributes.get("lineno", 1)
+        if isinstance(line_no_dummy, str):
+            line_no = int(line_no_dummy) if line_no_dummy else None
+        else:
+            line_no = line_no_dummy
         type = attributes.get("type", "KEYWORD")
         libname = attributes.get("libname")
         kwname = attributes.get("kwname")
@@ -978,7 +997,7 @@ class Debugger:
         handler: Any = None
         if type in ["KEYWORD", "SETUP", "TEARDOWN"]:
             try:
-                handler = EXECUTION_CONTEXTS.current.namespace.get_runner(name)._handler
+                handler = self.get_current_keyword_handler(name)
             except (SystemExit, KeyboardInterrupt):
                 raise
             except BaseException:
@@ -1009,6 +1028,7 @@ class Debugger:
         "BuiltIn.Run Keyword And Warn On Failure",
         "BuiltIn.Wait Until Keyword Succeeds",
         "BuiltIn.Run Keyword And Continue On Failure",
+        "BuiltIn.Run Keyword And Return Status",
     ]
 
     def is_not_caughted_by_keyword(self) -> bool:
@@ -1060,18 +1080,33 @@ class Debugger:
 
         return False
 
+    if get_robot_version() >= (7, 0):
+
+        def _get_step_data(self, step: Any) -> Any:
+            return step
+
+    else:
+
+        def _get_step_data(self, step: Any) -> Any:
+            return step.data
+
     def is_not_caugthed_by_except(self, message: Optional[str]) -> bool:
         if not message:
             return True
 
-        if get_robot_version() >= (5, 0) and self.debug_logger:
-            from robot.running.model import Try
+        if self.debug_logger:
+            if get_robot_version() >= (5, 0):
+                from robot.running.model import Try
 
-            if self.debug_logger.steps:
-                for branch in [f.data for f in reversed(self.debug_logger.steps) if isinstance(f.data, Try)]:
-                    for except_branch in branch.except_branches:
-                        if self._should_run_except(except_branch, message):
-                            return False
+                if self.debug_logger.steps:
+                    for branch in [
+                        self._get_step_data(f)
+                        for f in reversed(self.debug_logger.steps)
+                        if isinstance(self._get_step_data(f), Try)
+                    ]:
+                        for except_branch in branch.except_branches:
+                            if self._should_run_except(except_branch, message):
+                                return False
         return True
 
     def end_keyword(self, name: str, attributes: Dict[str, Any]) -> None:
@@ -1105,7 +1140,7 @@ class Debugger:
         handler: Any = None
         if type in ["KEYWORD", "SETUP", "TEARDOWN"]:
             try:
-                handler = EXECUTION_CONTEXTS.current.namespace.get_runner(name)._handler
+                handler = self.get_current_keyword_handler(name)
             except (SystemExit, KeyboardInterrupt):
                 raise
             except BaseException:
@@ -1369,6 +1404,16 @@ class Debugger:
 
         return Variable(name=name, value=repr(value), type=repr(type(value)))
 
+    if get_robot_version() >= (7, 0):
+
+        def get_handler_args(self, handler: UserKeywordHandler) -> Any:
+            return handler.args
+
+    else:
+
+        def get_handler_args(self, handler: UserKeywordHandler) -> Any:
+            return handler.arguments
+
     def get_variables(
         self,
         variables_reference: int,
@@ -1436,8 +1481,8 @@ class Debugger:
                                 }
                             )
 
-                            if entry.handler is not None and entry.handler.arguments:
-                                for argument in entry.handler.arguments.argument_names:
+                            if entry.handler is not None and self.get_handler_args(entry.handler):
+                                for argument in self.get_handler_args(entry.handler).argument_names:
                                     name = f"${{{argument}}}"
                                     try:
                                         value = vars[name]
@@ -1470,8 +1515,11 @@ class Debugger:
 
             if value is not None:
                 c = 0
+
+                padding = len(str(len(value)))
+
                 for i, v in enumerate(value, start or 0):
-                    result[str(i)] = self._create_variable(str(i), v)
+                    result[str(i)] = self._create_variable(str(i).zfill(padding), v)
                     c += 1
                     if count is not None and c >= count:
                         break
@@ -1494,6 +1542,16 @@ class Debugger:
     SPLIT_LINE: ClassVar = re.compile(r"(?= {2,}| ?\t)\s*")
     CURRDIR: ClassVar = re.compile(r"(?i)\$\{CURDIR\}")
 
+    if get_robot_version() >= (7, 0):
+
+        def _run_keyword(self, kw: Keyword, context: Any) -> Any:
+            return kw.run(context.steps[-1][1], context)
+
+    else:
+
+        def _run_keyword(self, kw: Keyword, context: Any) -> Any:
+            return kw.run(context)
+
     def evaluate(
         self,
         expression: str,
@@ -1519,6 +1577,9 @@ class Debugger:
         if evaluate_context is None:
             evaluate_context = EXECUTION_CONTEXTS.current
 
+        if stack_frame is None and evaluate_context is None:
+            return EvaluateResult(result="Unable to evaluate expression. No context available.", type="FatalError")
+
         result: Any = None
         try:
             if stack_frame is not None and stack_frame.source is not None:
@@ -1557,7 +1618,7 @@ class Debugger:
                                     args=tuple(splitted[1:]),
                                     assign=tuple(variables),
                                 )
-                                return kw.run(evaluate_context)
+                                return self._run_keyword(kw, evaluate_context)
 
                             result = self.run_in_robot_thread(run_kw)
 
@@ -1611,7 +1672,7 @@ class Debugger:
                             for kw in test.body:
                                 with LOGGER.delayed_logging:
                                     try:
-                                        result = kw.run(evaluate_context)
+                                        result = self._run_keyword(kw, evaluate_context)
                                     except (SystemExit, KeyboardInterrupt):
                                         raise
                                     except BaseException as e:
@@ -1793,6 +1854,22 @@ class Debugger:
 
         return result or None
 
+    if get_robot_version() >= (7, 0):
+
+        def _get_keywords_from_lib(self, lib: Any) -> Any:
+            return lib.keywords
+
+        def _get_short_doc_from_kw(self, kw: Any) -> Any:
+            return kw.short_doc
+
+    else:
+
+        def _get_keywords_from_lib(self, lib: Any) -> Any:
+            return lib.handlers
+
+        def _get_short_doc_from_kw(self, kw: Any) -> Any:
+            return kw.shortdoc
+
     def completions(
         self,
         text: str,
@@ -1824,14 +1901,14 @@ class Debugger:
                     type=CompletionItemType.MODULE,
                 )
             )
-            for kw in library.handlers:
+            for kw in self._get_keywords_from_lib(library):
                 result.append(
                     CompletionItem(
                         label=kw.name,
                         text=kw.name,
                         sort_text=f"001_{kw.name}",
                         type=CompletionItemType.FUNCTION,
-                        detail=kw.shortdoc,
+                        detail=self._get_short_doc_from_kw(kw),
                     )
                 )
 
@@ -1844,14 +1921,14 @@ class Debugger:
                     type=CompletionItemType.MODULE,
                 )
             )
-            for kw in resource.handlers:
+            for kw in self._get_keywords_from_lib(resource):
                 result.append(
                     CompletionItem(
                         label=kw.name,
                         text=kw.name,
                         sort_text=f"001_{kw.name}",
                         type=CompletionItemType.FUNCTION,
-                        detail=kw.shortdoc,
+                        detail=self._get_short_doc_from_kw(kw),
                     )
                 )
 
